/*
 * Copyright (C) 2001 TAKIZAWA Takashi <taki@cyber.email.ne.jp>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "cdb_make.h"
#include "env.h"
#include "exit.h"
#include "error.h"
#include "strerr.h"
#include "open.h"
#include "prot.h"

#include "auto_uids.h"
#include "auto_users.h"
#include "auto_qmailhome.h"

#include <pwd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#define FATAL "pwdbinit: fatal: "
#define WARN  "pwdbinit: "

static void die_write(char *dir,char *fn)
{
  strerr_die6sys(111,FATAL,"unable to write to authdb/",dir,"/",fn,": ");
}

struct passwd *get_pw(char *name)
{
  struct passwd *pw;
  pw = getpwnam(name);
  if (!pw) {
    if (errno == error_txtbsy)
      strerr_die2sys(111,FATAL,"unable to read /etc/passwd: ");
    else
      strerr_die4x(111,WARN,"user ",name," doesn't exist in /etc/passwd");
  }
  return pw;
}

static void setuidgid(int uid, int gid)
{
  if (prot_gid(gid) == -1)
    strerr_die4sys(111,FATAL,"unable to setgid to ",auto_vida,": ");
  if (prot_uid(uid) == -1)
    strerr_die4sys(111,FATAL,"unable to setuid to ",auto_authdb,": ");
}

static void authdbdirmake(uid_t owner,uid_t group)
{
  if (chdir(auto_qmailhome) == -1)
    strerr_die4sys(111,FATAL,"unable to chdir to ",auto_qmailhome,": ");
  if (chdir("authdb") == 0)
    return;
  if (mkdir("authdb",0700) == -1)
    strerr_die4sys(111,FATAL,"unable to mkdir ",auto_qmailhome,"/authdb: ");
  if (chown("authdb",owner,group) == -1)
    strerr_die6sys(111,FATAL,"unable to chown ",auto_qmailhome,"/authdb to",auto_authdb,": ");
  if (chdir("authdb") == -1)
    strerr_die4sys(111,FATAL,"unable to chdir to ",auto_qmailhome,"/authdb: ");
}

static void pwdbdirmake(char *pwdbdir)
{
  if (chdir(auto_qmailhome) == -1)
    strerr_die4sys(111,FATAL,"unable to chdir to ",auto_qmailhome,": ");
  if (chdir("authdb") == -1)
    strerr_die4sys(111,FATAL,"unable to chdir to ",auto_qmailhome,"/authdb: ");
  if (chdir(pwdbdir) == 0)
    strerr_die4x(111,WARN,"password database entry for ",pwdbdir," already exists");
  if (mkdir(pwdbdir,0700) == -1)
    strerr_die4sys(111,FATAL,"unable to mkdir authdb/",pwdbdir,": ");
  if (chdir(pwdbdir) == -1)
    strerr_die4sys(111,FATAL,"unable to chdir to authdb/",pwdbdir,": ");
}

static void makedb(char *pwdbdir)
{
  struct cdb_make c;
  int fd;

  fd = open_trunc("passwd");
  if (fd == -1)
    strerr_die4sys(111,FATAL,"unable to create authdb/",pwdbdir,"/passwd: ");
  if (write(fd,"#user\tpassword\tenviron\n",23) == -1)
    { close(fd); die_write(pwdbdir,"passwd"); }
  if (fsync(fd) == -1) die_write(pwdbdir,"passwd");
  if (close(fd) == -1) die_write(pwdbdir,"passwd");

  fd = open_trunc("cdb");
  if (fd == -1)
    strerr_die4sys(111,FATAL,"unable to create authdb/",pwdbdir,"/cdb: ");
  if (cdb_make_start(&c,fd) == -1) die_write(pwdbdir,"cdb");
  if (cdb_make_finish(&c) == -1) die_write(pwdbdir,"cdb");
  if (fsync(fd) == -1) die_write(pwdbdir,"cdb");
  if (close(fd) == -1) die_write(pwdbdir,"cdb");
}

int main(int argc,char **argv)
{
  char *domainowner;
  struct passwd *pw;
  uid_t pw_uid;

  pw_uid = getuid();
  if (pw_uid)
    strerr_die2x(111,FATAL,"you must run as root: ");

  umask(077);

  authdbdirmake(auto_uidad,auto_gidvida);
  setuidgid(auto_uidad,auto_gidvida);

  domainowner = argv[1];
  if (domainowner) {
    pw = get_pw(domainowner);
    pwdbdirmake(domainowner);
    makedb(domainowner);
  }
  else {
    pwdbdirmake(auto_realdomainowner);
    makedb(auto_realdomainowner);
    pwdbdirmake("pwdusers");
    makedb("pwdusers");
  }

  _exit(0);
}
